//: [Previous](@previous)

import Foundation

struct Entry: Decodable, CustomDebugStringConvertible {
    let text: String
    
    public var debugDescription: String {
        return "\(text)"
    }
}

import Foundation

struct TextEntries {
    let values: [Entry]
}

extension TextEntries: Decodable {
    init(from decoder: Decoder) throws {
        var entries = try decoder.unkeyedContainer()
        
        var values = [Entry]()
        while !entries.isAtEnd {
            let entry = try entries.decode(Entry.self)
            values.append(entry)
        }
        
        self.values = values
    }
}

struct Error: Codable {
    enum Key: String, CodingKey {
        case error
    }

    let code: Int
    let message: String
    
    init(from decoder: Decoder) throws {
        let response = try decoder.container(keyedBy: Key.self)
        let contents = try response.nestedContainer(keyedBy: CodingKeys.self, forKey: .error)
        
        self.code = try contents.decode(Int.self, forKey: .code)
        self.message = try contents.decode(String.self, forKey: .message)
    }
}

extension Error: Swift.Error, CustomNSError {
    public static var errorDomain: String = "com.qnoid"
    
    var errorCode: Int {
        return code
    }
    
    var errorUserInfo: [String : Any] {
        [NSLocalizedDescriptionKey : message]
    }
}

struct Contents {
    enum Key: String, CodingKey {
        case contents
    }
    enum CodingKeys: String, CodingKey {
        case entries = "text_entries"
    }

    let entries: TextEntries
}

extension Contents: Decodable {
    init(from decoder: Decoder) throws {
        let response = try decoder.container(keyedBy: Key.self)
        let contents = try response.nestedContainer(keyedBy: CodingKeys.self, forKey: .contents)
        
        self.entries = try contents.decode(TextEntries.self, forKey: .entries)
    }
}

var error = """
{
    "success": false,
    "error": {
        "code": 1,
        "message": "User not found"
    }
}
"""

var success = """
{
    "success": true,
    "contents": {
        "text_entries": [{
                "text": "Hello!",
            },
            {
                "text": "¡Hola!",
            }
        ]
    }
}
"""
let data = error.data(using: .utf8)!

let result = Result<Contents, Swift.Error> {
        do {
            return try JSONDecoder().decode(Contents.self, from: data)
        } catch {
            throw try JSONDecoder().decode(Error.self, from: data)
        }
    }

do {
    let contents = try result.get()
    contents.entries.values.map { entry in
        entry.text
    }
}
catch {
    error
}

class Response {

    private let data: Data?
    private let error: Swift.Error?
    
    var result: Result<Contents, Swift.Error> {
        
        switch (data, error) {
        case (_, let error?):
            return Result.failure(error)
        case (let data?, _):
            return Result<Contents, Swift.Error> {
                
                do {
                    return try JSONDecoder().decode(Contents.self, from: data)
                } catch {
                    throw try JSONDecoder().decode(Error.self, from: data)
                }
            }
        case (.none, .none):
            preconditionFailure("Must have data returned or an error")
        }
    }
    
    init(data: Data?, error: Swift.Error?) {
        self.data = data
        self.error = error
    }
}

let response = Response(data: success.data(using: .utf8), error: nil)
DispatchQueue.main.async { [result = response.result] in
    guard let content = try? result.get() else {
        //error handling
        return
    }
    content.entries.values.map { entry in
        entry.text
    }
}
